import pytest
from pathlib import Path
from Z0Z_tools import pipAnything
import sys
import shutil

TEST_DATA_DIR = Path('tests/dataSamples/tmp')

@pytest.fixture(autouse=True)
def setup_test_directory():
    """Ensure test directory exists and is clean before each test."""
    TEST_DATA_DIR.mkdir(parents=True, exist_ok=True)
    yield
    if TEST_DATA_DIR.exists():
        shutil.rmtree(TEST_DATA_DIR)

@pytest.fixture
def requirements_file():
    """Create a requirements file in the test directory."""
    req_file = TEST_DATA_DIR / 'requirements.txt'
    req_file.write_text("""
    # This is a comment
    package-A==1.2.3
    package-B>=4.5.6,<=7.8.9
    package_C
    # Another comment
    analyzeAudio@git+https://github.com/hunterhogan/analyzeAudio.git
    """)
    return req_file

@pytest.fixture
def package_dir():
    """Create a fake package directory with requirements."""
    pkg_dir = TEST_DATA_DIR / 'test_package'
    pkg_dir.mkdir(exist_ok=True)
    (pkg_dir / 'requirements.txt').write_text('numpy\npandas')
    return pkg_dir

def test_makeListRequirementsFromRequirementsFile(requirements_file):
    """Test requirements file parsing with various inputs."""
    requirements = pipAnything.makeListRequirementsFromRequirementsFile(requirements_file)
    assert len(requirements) == 4
    assert 'package-A==1.2.3' in requirements
    assert 'package-B>=4.5.6,<=7.8.9' in requirements
    assert 'package_C' in requirements
    assert 'analyzeAudio@git+https://github.com/hunterhogan/analyzeAudio.git' in requirements

def test_multiple_requirements_files():
    """Test processing multiple requirements files."""
    req1 = TEST_DATA_DIR / 'req1.txt'
    req2 = TEST_DATA_DIR / 'req2.txt'
    
    req1.write_text('package-A==1.0\npackage-B==2.0')
    req2.write_text('package-B==2.0\npackage-C==3.0')

    requirements = pipAnything.makeListRequirementsFromRequirementsFile(req1, req2)
    assert len(requirements) == 3
    assert sorted(requirements) == ['package-A==1.0', 'package-B==2.0', 'package-C==3.0']

def test_nonexistent_requirements_file():
    """Test handling of non-existent requirements file."""
    nonexistent = TEST_DATA_DIR / 'nonexistent.txt'
    requirements = pipAnything.makeListRequirementsFromRequirementsFile(nonexistent)
    assert len(requirements) == 0

@pytest.mark.parametrize("content,expected_count", [
    ("invalid==requirement==1.0\nvalid-package==1.0", 1),
    ("spaces in package==1.0\n@#$%^invalid\nvalid-pkg==1.0", 1),
    ("valid-1==1.0\nvalid-2==2.0", 2),
])
def test_invalid_requirements(content, expected_count):
    """Test handling of invalid requirements content."""
    req_file = TEST_DATA_DIR / 'requirements.txt'
    req_file.write_text(content)
    requirements = pipAnything.makeListRequirementsFromRequirementsFile(req_file)
    assert len(requirements) == expected_count

def test_make_setupDOTpy():
    """Test setup.py content generation."""
    relative_path = 'my_package'
    requirements = ['numpy', 'pandas']
    setup_content = pipAnything.make_setupDOTpy(relative_path, requirements)
    
    assert f"name='{Path(relative_path).name}'" in setup_content
    assert f"packages=find_packages(where=r'{relative_path}')" in setup_content
    assert f"package_dir={{'': r'{relative_path}'}}" in setup_content
    assert f"install_requires={requirements}," in setup_content
    assert "include_package_data=True" in setup_content

def test_installPackageTarget(mocker, package_dir):
    """Test package installation process."""
    mock_popen = mocker.patch('subprocess.Popen')
    mock_process = mocker.MagicMock()
    mock_process.stdout = ['Installing...', 'Done!']
    mock_process.wait.return_value = 0
    mock_popen.return_value = mock_process
    
    # No need to mock mkdtemp anymore since we're using TEST_DATA_DIR
    
    pipAnything.installPackageTarget(package_dir)
    
    mock_popen.assert_called_once()
    args = mock_popen.call_args[1]['args']
    assert args[0] == sys.executable
    assert args[1:4] == ['-m', 'pip', 'install']

@pytest.mark.parametrize("argv,should_exit", [
    (['script.py'], True),
    (['script.py', '/nonexistent/path'], True),
])
def test_CLI_functions(mocker, argv, should_exit):
    """Test CLI argument handling."""
    mocker.patch('sys.argv', argv)
    mock_exit = mocker.patch('sys.exit')
    mock_print = mocker.patch('builtins.print')
    
    pipAnything.everyone_knows_what___main___is()
    
    if should_exit:
        mock_exit.assert_called_once_with(1)
        mock_print.assert_called()

def test_snark_level(mocker):
    """Test appropriate levels of snark in messages."""
    mocker.patch('sys.argv', ['script.py'])
    mock_exit = mocker.patch('sys.exit')
    mock_print = mocker.patch('builtins.print')
    
    pipAnything.main()
    
    mock_print.assert_called()
    mock_exit.assert_called_once_with(1)
    
    printed_messages = ' '.join(str(call.args[0]) 
                              for call in mock_print.call_args_list)
    assert 'obviously' not in printed_messages.lower()


from Z0Z_tools.dataStructures import stringItUp, updateExtendPolishDictionaryLists
from Z0Z_tools.ioAudio import writeWav, readAudioFile, loadWaveforms
from Z0Z_tools.parseParameters import defineConcurrencyLimit, oopsieKwargsie, intInnit
from Z0Z_tools.pipAnything import installPackageTarget, makeListRequirementsFromRequirementsFile
from Z0Z_tools.Z0Z_io import dataTabularTOpathFilenameDelimited, findRelativePath

__all__ = [
    'dataTabularTOpathFilenameDelimited',
    'defineConcurrencyLimit',
    'findRelativePath',
    'installPackageTarget',
    'intInnit',
    'loadWaveforms',
    'makeListRequirementsFromRequirementsFile',
    'oopsieKwargsie',
    'readAudioFile',
    'stringItUp',
    'updateExtendPolishDictionaryLists',
    'writeWav',
]




import multiprocessing
from typing import Any, Callable, List, Optional, Union
from unittest.mock import patch


def defineConcurrencyLimit(limit: Optional[Union[int, float, bool]]) -> int:
    """
    Determine the concurrency limit based on the provided `limit` parameter.

    Parameters:
        limit: The concurrency limit specification. Your user can set a limit with 
        a simple True/False, a quantity of CPUs, or a ratio of CPUs. Furthermore, positive
        numbers define the maximum usage, and negative numbers define the amount to keep
        in reserve.

    Returns:
        concurrencyLimit: The calculated concurrency limit, ensuring it is at least 1.

    If you want to be extra nice to your users, consider using `oopsieKwargsie()` to handle
    malformed inputs.
    """
    """Example docstring:
    Parameters:
        CPUlimit: whether and how to limit the CPU usage. See notes for details. 

    Limits on CPU usage `CPUlimit`:
        - `False`, `None`, or `0`: No limits on CPU usage; uses all available CPUs. All other values will potentially limit CPU usage.
        - `True`: Yes, limit the CPU usage; limits to 1 CPU.
        - Integer `>= 1`: Limits usage to the specified number of CPUs.
        - Decimal value (`float`) between 0 and 1: Fraction of total CPUs to use.
        - Decimal value (`float`) between -1 and 0: Fraction of CPUs to *not* use.
        - Integer `<= -1`: Subtract the absolute value from total CPUs.
    """
    cpuTotal = multiprocessing.cpu_count()
    concurrencyLimit = cpuTotal

    if isinstance(limit, str):
        limit = oopsieKwargsie(limit) # type: ignore

    match limit:
        case None | False | 0:
            pass
        case True:
            concurrencyLimit = 1
        case _ if limit >= 1:
            concurrencyLimit = limit
        case _ if 0 < limit < 1:
            concurrencyLimit = int(limit * cpuTotal)
        case _ if -1 < limit < 0:
            concurrencyLimit = cpuTotal - abs(int(limit * cpuTotal))
        case _ if limit <= 1:
            concurrencyLimit = cpuTotal - abs(limit)

    return max(int(concurrencyLimit), 1)

def oopsieKwargsie(huh: str) -> None | str | bool:
    """
    If a calling function passes a `str` to a parameter that shouldn't receive a `str`, `oopsieKwargsie()` might help you avoid an Exception. It tries to interpret the string as `True`, `False`, or `None`.

    Parameters:
        huh: The input string to be parsed.

    Returns:
        (bool | None | str): The reserved keywords `True`, `False`, or `None` or the original string, `huh`.
    """
    formatted = huh.strip().title()
    if formatted == str(True):
        return True
    elif formatted == str(False):
        return False
    elif formatted == str(None):
        return None
    else:
        return huh

def intInnit(listInt_Allegedly: List[int], parameterName: str = 'unnamed parameter') -> List[int]:
    """
    Rigorously validates and converts input to a list of non-negative integers.
    Also returns a filtered list containing only positive integers.

    Parameters:
        listInt_Allegedly: Input that should be a list of integers
        parameterName: Name of parameter for error messages

    Returns:
        listValidated: List of integers as `int` type

    Raises:
        Various built-in Python exceptions with enhanced error messages.
    """
    if not listInt_Allegedly:
        raise ValueError(f"{parameterName} must not be empty")

    try:
        iter(listInt_Allegedly)
        lengthInitial = len(listInt_Allegedly)
        
        listValidated = []
        for allegedInt in listInt_Allegedly:
            if isinstance(allegedInt, bool):
                raise TypeError(f"Boolean values ({allegedInt}) are not integers.")
            elif isinstance(allegedInt, (int, float)):
                if float(allegedInt).is_integer():
                    allegedInt = int(allegedInt)
                else:
                    raise ValueError(f"Value {allegedInt} must be a whole number")
            else:
                raise TypeError(f"Value {allegedInt} must be numeric")
            
            listValidated.append(allegedInt)
                
            if len(listInt_Allegedly) != lengthInitial:
                raise RuntimeError("Input sequence was modified during iteration")

        return listValidated

    except TypeError as ERRORtype:
        if not hasattr(listInt_Allegedly, '__iter__'):
            ERRORmessage = f"{parameterName} does not have the '__iter__' attribute (it is not iterable), but it must have the '__iter__' attribute. Value was passed as data type '{type(listInt_Allegedly)}'."
        else:
            ERRORmessage = f"Invalid element in {parameterName}: {ERRORtype.args[0]}"
        raise TypeError(ERRORmessage) from None

if __name__ == '__main__':
    multiprocessing.set_start_method('spawn')



"""
Functions:
    - installPackageTarget: Tries to trick pip into installing the package from a given directory.
    - makeListRequirementsFromRequirementsFile: Reads a requirements.txt file, discards anything it couldn't understand, and creates a list of packages. 

Usage:
    from pipAnything import installPackageTarget
    installPackageTarget('path/to/packageTarget')

    pip will attempt to install requirements.txt, but don't rely on dependencies being installed.
"""

from packaging.requirements import Requirement
from typing import List, Union
import os
import pathlib
import subprocess
import sys
import tempfile
from Z0Z_tools.Z0Z_io import findRelativePath

def makeListRequirementsFromRequirementsFile(*pathFilenames: Union[str, os.PathLike[str]]) -> List[str]:
    """
    Reads one or more requirements files and extracts valid package requirements.
    Parameters:
        *pathFilenames: One or more paths to requirements files.
    Returns:
        listRequirements: A list of unique, valid package requirements found in the provided files.
    The function performs the following steps:
    1. Iterates over each provided file path.
    2. Checks if the file exists.
    3. Reads the file line by line, removing comments and trimming whitespace.
    4. Skips lines that are empty or contain spaces/tabs after sanitization.
    5. Validates if the sanitized line is a valid requirement.
    6. Collects valid requirements and removes duplicates before returning the list.
    """
    listRequirements = []
    
    for pathFilename in pathFilenames:
        if pathlib.Path(pathFilename).exists():
            try:
                filesystemObjectRead = open(pathFilename, 'r')
                for commentedLine in filesystemObjectRead:
                    sanitizedLine = commentedLine.split('#')[0].strip()  # Remove comments and trim whitespace

                    # Skip lines that are empty or contain spaces/tabs after sanitization
                    if "\t" in sanitizedLine or " " in sanitizedLine or not sanitizedLine:
                        continue

                    # Validate if it's a valid requirement
                    try:
                        Requirement(sanitizedLine)
                        listRequirements.append(sanitizedLine)
                    except:
                        pass  # Skip invalid requirement lines
            finally:
                filesystemObjectRead.close()

    return sorted(set(listRequirements))  # Remove duplicates

def make_setupDOTpy(relativePathPackage: Union[str, os.PathLike[str]], listRequirements: List[str]) -> str:
    """
    Generates setup.py file content for installing the package.

    Parameters:
        relativePathPackage: The relative path to the package directory.
        listRequirements: A list of requirements to be included in install_requires.

    Returns:
        setupDOTpy: The setup.py content to be written to a file.
    """
    return rf"""
import os
from setuptools import setup, find_packages

setup(
    name='{pathlib.Path(relativePathPackage).name}',
    version='0.0.0',
    packages=find_packages(where=r'{relativePathPackage}'),
    package_dir={{'': r'{relativePathPackage}'}},
    install_requires={listRequirements},
    include_package_data=True,
)
""" 

def installPackageTarget(pathPackageTarget: Union[str, os.PathLike[str]]) -> None:
    """
    Installs a package by creating a temporary setup.py and tricking pip into installing it.

    Parameters:
        pathPackageTarget: The directory path of the package to be installed.
    """
    filenameRequirementsHARDCODED = pathlib.Path('requirements.txt')
    filenameRequirements = pathlib.Path(filenameRequirementsHARDCODED)

    pathPackage = pathlib.Path(pathPackageTarget).resolve()
    pathSystemTemporary = pathlib.Path(tempfile.mkdtemp())
    pathFilename_setupDOTpy = pathSystemTemporary / 'setup.py'

    pathFilenameRequirements = pathPackage / filenameRequirements
    listRequirements = makeListRequirementsFromRequirementsFile(pathFilenameRequirements)

    # Try-finally block for file handling: with-as doesn't always work
    writeStream = None
    try:
        writeStream = pathFilename_setupDOTpy.open(mode='w')
        relativePathPackage = findRelativePath(pathSystemTemporary, pathPackage)
        writeStream.write(make_setupDOTpy(relativePathPackage, listRequirements))
    finally:
        if writeStream:
            writeStream.close()

    # Run pip to install the package from the temporary directory
    subprocessPython = subprocess.Popen(
    # `pip` needs a RELATIVE PATH, not an absolute path, and not a path+filename. 
        args=[sys.executable, '-m', 'pip', 'install', str(pathSystemTemporary)],
        stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True
    )

    # Output the subprocess stdout in real-time
    if subprocessPython.stdout:
        for lineStdout in subprocessPython.stdout:
            print(lineStdout, end="")

    subprocessPython.wait()

    # Clean up by removing setup.py
    pathFilename_setupDOTpy.unlink()

def everyone_knows_what___main___is() -> None:
    """A rudimentary CLI for the module.
    call `installPackageTarget` from other modules."""
    packageTarget = sys.argv[1] if len(sys.argv) > 1 else ''
    pathPackageTarget = pathlib.Path(packageTarget)
    if not pathPackageTarget.is_dir() or len(sys.argv) != 2:
        namespaceModule = pathlib.Path(__file__).stem
        namespacePackage = pathlib.Path(__file__).parent.stem
        print(f"\n{namespaceModule} says, 'That didn't work. Try again?'\n\n"
              f"Usage:\tpython -m {namespacePackage}.{namespaceModule} <packageTarget>\n"
              f"\t<packageTarget> is a path to a directory with Python modules\n"
              f"\tExample: python -m {namespacePackage}.{namespaceModule} '{pathlib.PurePath('path' ,'to', 'Z0Z_tools')}'") 
        # What is `-m`? Obviously, `-m` creates a namespace for the module, which is obviously necessary, except when it isn't.
        sys.exit(1)

    installPackageTarget(pathPackageTarget)
    print(f"\n{pathlib.Path(__file__).stem} finished trying to trick pip into installing {pathPackageTarget.name}. Did it work?")

def readability_counts() -> None:
    """Brings the snark."""
    everyone_knows_what___main___is()

def main() -> None:
    """Jabs subtly."""
    readability_counts()

if __name__ == "__main__":
    main()


from typing import Optional, Union, Callable, Any
import pytest
from unittest.mock import patch
from Z0Z_tools import defineConcurrencyLimit, oopsieKwargsie

@pytest.fixture
def mockCpuCount():
    """Fixture to mock multiprocessing.cpu_count(). Always returns 8."""
    with patch('multiprocessing.cpu_count', return_value=8):
        yield

@pytest.mark.parametrize("variantTrue", ['True', 'TRUE', ' true ', 'TrUe'])
def testOopsieKwargsieHandlesTrueVariants(variantTrue):
    """Test oopsieKwargsie handles various string representations of True."""
    assert oopsieKwargsie(variantTrue) is True

@pytest.mark.parametrize("variantFalse", ['False', 'FALSE', ' false ', 'FaLsE'])
def testOopsieKwargsieHandlesFalseVariants(variantFalse):
    """Test oopsieKwargsie handles various string representations of False."""
    assert oopsieKwargsie(variantFalse) is False

@pytest.mark.parametrize("variantNone", ['None', 'NONE', ' none ', 'NoNe'])
def testOopsieKwargsieHandlesNoneVariants(variantNone):
    """Test oopsieKwargsie handles various string representations of None."""
    assert oopsieKwargsie(variantNone) is None

@pytest.mark.parametrize("stringInput", ['hello', '123', 'True story', 'False alarm'])
def testOopsieKwargsieReturnsOriginalString(stringInput):
    """Test oopsieKwargsie returns the original string when input is unrecognized."""
    assert oopsieKwargsie(stringInput) == stringInput

@pytest.mark.usefixtures("mockCpuCount")
@pytest.mark.parametrize("limitParameter, expectedLimit", [
    (None, 8),
    (False, 8),
    (0, 8),
])
def testDefineConcurrencyLimitDefaults(limitParameter, expectedLimit):
    """Test defineConcurrencyLimit with default parameters."""
    resultLimit = defineConcurrencyLimit(limitParameter)
    assert resultLimit == expectedLimit

@pytest.mark.usefixtures("mockCpuCount")
@pytest.mark.parametrize("limitParameter, expectedLimit", [
    (1, 1),
    (4, 4),
    (16, 16),
])
def testDefineConcurrencyLimitDirectIntegers(limitParameter, expectedLimit):
    """Test defineConcurrencyLimit with direct integer values ≥ 1."""
    resultLimit = defineConcurrencyLimit(limitParameter)
    assert resultLimit == expectedLimit

@pytest.mark.usefixtures("mockCpuCount")
@pytest.mark.parametrize("limitParameter, expectedLimit", [
    (0.5, 4),
    (0.25, 2),
    (0.75, 6),
])
def testDefineConcurrencyLimitFractionalFloats(limitParameter, expectedLimit):
    """Test defineConcurrencyLimit with float values between 0 and 1."""
    resultLimit = defineConcurrencyLimit(limitParameter)
    assert resultLimit == expectedLimit

@pytest.mark.usefixtures("mockCpuCount")
@pytest.mark.parametrize("limitParameter, expectedLimit", [
    (-0.5, 4),
    (-0.25, 6),
    (-0.75, 2),
])
def testDefineConcurrencyLimitNegativeFractions(limitParameter, expectedLimit):
    """Test defineConcurrencyLimit with float values between -1 and 0."""
    resultLimit = defineConcurrencyLimit(limitParameter)
    assert resultLimit == expectedLimit

@pytest.mark.usefixtures("mockCpuCount")
@pytest.mark.parametrize("limitParameter, expectedLimit", [
    (-1, 7),
    (-3, 5),
    (-7, 1),
])
def testDefineConcurrencyLimitNegativeIntegers(limitParameter, expectedLimit):
    """Test defineConcurrencyLimit with integer values ≤ -1."""
    resultLimit = defineConcurrencyLimit(limitParameter)
    assert resultLimit == expectedLimit

@pytest.mark.usefixtures("mockCpuCount")
def testDefineConcurrencyLimitBooleanTrue():
    """Test defineConcurrencyLimit with boolean True."""
    resultLimit = defineConcurrencyLimit(True)
    assert resultLimit == 1

@pytest.mark.usefixtures("mockCpuCount")
def testDefineConcurrencyLimitSpecificTrueCase():
    """Ensure defineConcurrencyLimit(True) specifically returns 1."""
    resultTrue = defineConcurrencyLimit(True)
    resultNone = defineConcurrencyLimit(None)
    assert resultTrue == 1
    assert resultNone != 1

@pytest.mark.usefixtures("mockCpuCount")
@patch('Z0Z_tools.parseParameters.oopsieKwargsie')
@pytest.mark.parametrize("stringInput, mockedReturn, expectedLimit", [
    ("true", True, 1),
    ("false", False, 8),
    ("none", None, 8),
    ("4", 4, 4),
])
def testDefineConcurrencyLimitStringValues(mockOopsieKwargsie, stringInput, mockedReturn, expectedLimit):
    """Test defineConcurrencyLimit handling of string inputs via oopsieKwargsie."""
    mockOopsieKwargsie.return_value = mockedReturn
    resultLimit = defineConcurrencyLimit(stringInput)
    assert resultLimit == expectedLimit

@pytest.mark.usefixtures("mockCpuCount")
@pytest.mark.parametrize("limitParameter, expectedLimit", [
    (-10, 1),
    (-0.99, 1),
    (0.1, 1),
])
def testDefineConcurrencyLimitEnsuresMinimumOne(limitParameter, expectedLimit):
    """Test defineConcurrencyLimit ensures the minimum return value is 1."""
    resultLimit = defineConcurrencyLimit(limitParameter)
    assert resultLimit == expectedLimit

def makeTestSuiteOopsieKwargsie(functionUnderTest: Callable[[str], Optional[Union[bool, str]]]):
    """
    Creates a test suite for oopsieKwargsie-like functions.
    
    Parameters:
        functionUnderTest: The function to test, must accept str and return bool|None|str

    Returns:
        dictionaryTests: Dictionary of test functions to run
    """
    def testHandlesTrueVariants():
        for variantTrue in ['True', 'TRUE', ' true ', 'TrUe']:
            assert functionUnderTest(variantTrue) is True

    def testHandlesFalseVariants():
        for variantFalse in ['False', 'FALSE', ' false ', 'FaLsE']:
            assert functionUnderTest(variantFalse) is False

    def testHandlesNoneVariants():
        for variantNone in ['None', 'NONE', ' none ', 'NoNe']:
            assert functionUnderTest(variantNone) is None

    def testReturnsOriginalString():
        for stringInput in ['hello', '123', 'True story', 'False alarm']:
            assert functionUnderTest(stringInput) == stringInput

    return {
        'testHandlesTrueVariants': testHandlesTrueVariants,
        'testHandlesFalseVariants': testHandlesFalseVariants,
        'testHandlesNoneVariants': testHandlesNoneVariants,
        'testReturnsOriginalString': testReturnsOriginalString
    }

def makeTestSuiteConcurrencyLimit(functionUnderTest: Callable[[Any], int], cpuCount: int = 8):
    """
    Creates a test suite for defineConcurrencyLimit-like functions.
    
    Parameters:
        functionUnderTest: The function to test, must return int
        cpuCount (8): Number of CPUs to simulate

    Returns:
        dictionaryTests: Dictionary of test functions to run
    """
    @patch('multiprocessing.cpu_count', return_value=cpuCount)
    def testDefaults(_mockCpu):
        for limitParameter in [None, False, 0]:
            assert functionUnderTest(limitParameter) == cpuCount

    @patch('multiprocessing.cpu_count', return_value=cpuCount)
    def testDirectIntegers(_mockCpu):
        for limitParameter in [1, 4, 16]:
            assert functionUnderTest(limitParameter) == limitParameter

    @patch('multiprocessing.cpu_count', return_value=cpuCount)
    def testFractionalFloats(_mockCpu):
        testCases = {
            0.5: cpuCount // 2,
            0.25: cpuCount // 4,
            0.75: int(cpuCount * 0.75)
        }
        for input, expected in testCases.items():
            assert functionUnderTest(input) == expected

    @patch('multiprocessing.cpu_count', return_value=cpuCount)
    def testMinimumOne(_mockCpu):
        for limitParameter in [-10, -0.99, 0.1]:
            assert functionUnderTest(limitParameter) >= 1

    return {
        'testDefaults': testDefaults,
        'testDirectIntegers': testDirectIntegers,
        'testFractionalFloats': testFractionalFloats,
        'testMinimumOne': testMinimumOne
    }

def makeTestSuiteIntInnit(functionUnderTest):
    """
    Creates a test suite for intInnit-like functions.
    
    Parameters:
        functionUnderTest: The function to test, must accept list and return list[int]

    Returns:
        dictionaryTests: Dictionary of test functions to run
    """
    def testHandlesValidIntegers():
        assert functionUnderTest([1, 2, 3], 'test') == [1, 2, 3]
        assert functionUnderTest([1.0, 2.0, 3.0], 'test') == [1, 2, 3]

    def testRejectsNonWholeNumbers():
        with pytest.raises(ValueError):
            functionUnderTest([1.5, 2, 3], 'test')

    def testRejectsBooleans():
        with pytest.raises(TypeError):
            functionUnderTest([True, False], 'test')

    def testRejectsNonNumerics():
        with pytest.raises(TypeError):
            functionUnderTest(['1', '2', '3'], 'test')

    def testRejectsEmptyList():
        with pytest.raises(ValueError):
            functionUnderTest([], 'test')

    def testHandlesMixedValidTypes():
        assert functionUnderTest([1, 2.0, 3], 'test') == [1, 2, 3]

    return {
        'testHandlesValidIntegers': testHandlesValidIntegers,
        'testRejectsNonWholeNumbers': testRejectsNonWholeNumbers,
        'testRejectsBooleans': testRejectsBooleans,
        'testRejectsNonNumerics': testRejectsNonNumerics,
        'testRejectsEmptyList': testRejectsEmptyList,
        'testHandlesMixedValidTypes': testHandlesMixedValidTypes
    }

@pytest.mark.parametrize("input_list,expected", [
    ([1, 2, 3], [1, 2, 3]),
    ([1.0, 2.0, 3.0], [1, 2, 3]),
    ([1, 2.0, 3], [1, 2, 3]),
])
def testIntInnitValidCases(input_list, expected):
    """Test intInnit with valid inputs."""
    from Z0Z_tools import intInnit
    assert intInnit(input_list, 'test') == expected

@pytest.mark.parametrize("invalid_input,expected_error", [
    ([1.5, 2, 3], ValueError),
    ([True, False], TypeError),
    (['1', '2', '3'], TypeError),
    ([], ValueError),
])
def testIntInnitInvalidCases(invalid_input, expected_error):
    """Test intInnit with invalid inputs."""
    from Z0Z_tools import intInnit
    with pytest.raises(expected_error):
        intInnit(invalid_input, 'test')


import io
import os
import pathlib
from typing import Any, BinaryIO, Dict, List, Sequence, Tuple, Union

import numpy
import resampy
import soundfile
from numpy.typing import NDArray


def loadWaveforms(listPathFilenames: Union[Sequence[str], Sequence[os.PathLike[str]]], sampleRate: int = 44100) -> NDArray[numpy.float32]:
    """
    Load a list of audio files into a single array.
    Parameters:
        listPathFilenames: List of file paths to the audio files.
        sampleRate (44100): Target sample rate for the waveforms; the function will resample if necessary. Defaults to 44100.
    Returns:
        arrayWaveforms: A single NumPy array of shape (COUNTchannels, COUNTsamplesMaximum, COUNTwaveforms)
    """
    axisOrderMapping: Dict[str, int] = {'indexingAxis': -1, 'axisTime': -2, 'axisChannels': 0}
    axesSizes: Dict[str, int] = {keyName: 1 for keyName in axisOrderMapping.keys()}
    COUNTaxes: int = len(axisOrderMapping)
    listShapeIndexToSize: List[int] = [9001] * COUNTaxes

    COUNTwaveforms: int = len(listPathFilenames)
    axesSizes['indexingAxis'] = COUNTwaveforms
    COUNTchannels: int = 2
    axesSizes['axisChannels'] = COUNTchannels

    listCOUNTsamples: List[int] = []
    axisTime: int = 0
    for pathFilename in listPathFilenames:
        try:
            with soundfile.SoundFile(pathFilename) as readSoundFile:
                sampleRateSoundFile: int = readSoundFile.samplerate
                waveform: NDArray[numpy.float32] = readSoundFile.read(dtype='float32', always_2d=True).astype(numpy.float32)
                if sampleRateSoundFile != sampleRate:
                    waveform = resampleWaveform(waveform, sampleRate, sampleRateSoundFile)
                listCOUNTsamples.append(waveform.shape[axisTime])
        except soundfile.LibsndfileError as ERRORmessage:
            if 'System error' in str(ERRORmessage):
                raise FileNotFoundError(f"File not found: {pathFilename}") from ERRORmessage
            else:
                raise
            
    COUNTsamplesMaximum: int = max(listCOUNTsamples)
    axesSizes['axisTime'] = COUNTsamplesMaximum

    for keyName, axisSize in axesSizes.items():
        axisNormalized: int = (axisOrderMapping[keyName] + COUNTaxes) % COUNTaxes
        listShapeIndexToSize[axisNormalized] = axisSize
    tupleShapeArray: Tuple[int, ...] = tuple(listShapeIndexToSize)

    # `numpy.zeros` so that shorter waveforms are safely padded with zeros
    arrayWaveforms: NDArray[numpy.float32] = numpy.zeros(tupleShapeArray, dtype=numpy.float32)

    for index in range(COUNTwaveforms):
        with soundfile.SoundFile(listPathFilenames[index]) as readSoundFile:
            sampleRateSoundFile: int = readSoundFile.samplerate
            waveform: NDArray[numpy.float32] = readSoundFile.read(dtype='float32', always_2d=True).astype(numpy.float32)

            if sampleRateSoundFile != sampleRate:
                waveform = resampleWaveform(waveform, sampleRate, sampleRateSoundFile)

            COUNTsamples: int = waveform.shape[axisTime]
            arrayWaveforms[:, 0:COUNTsamples, index] = waveform.T

    return arrayWaveforms

def readAudioFile(pathFilename: Union[str, os.PathLike[Any], BinaryIO], sampleRate: int = 44100) -> NDArray[numpy.float32]:
    """
    Reads an audio file and returns its data as a NumPy array. Mono is always converted to stereo.

    Parameters:
        pathFilename: The path to the audio file.
        sampleRate (44100): The sample rate to use when reading the file. Defaults to 44100.

    Returns:
        waveform: The audio data in an array shaped (channels, samples).
    """
    try:
        with soundfile.SoundFile(pathFilename) as readSoundFile:
            sampleRateSource: int = readSoundFile.samplerate
            waveform: NDArray[numpy.float32] = readSoundFile.read(dtype='float32', always_2d=True).astype(numpy.float32)
            waveform = resampleWaveform(waveform, sampleRateDesired=sampleRate, sampleRateSource=sampleRateSource)
            # If the audio is mono (1 channel), convert it to stereo by duplicating the channel
            if waveform.shape[1] == 1:
                waveform = numpy.repeat(waveform, 2, axis=1)
            return waveform.T
    except soundfile.LibsndfileError as ERRORmessage:
        if 'System error' in str(ERRORmessage):
            raise FileNotFoundError(f"File not found: {pathFilename}") from ERRORmessage
        else:
            raise

def resampleWaveform(waveform: NDArray[numpy.float32], sampleRateDesired: int, sampleRateSource: int) -> NDArray[numpy.float32]:
    """
    Resamples the waveform to the desired sample rate using resampy.

    Parameters:
        waveform: The input audio data.
        sampleRateDesired: The desired sample rate.
        sampleRateSource: The original sample rate of the waveform.

    Returns:
        waveformResampled: The resampled waveform.
    """
    if sampleRateSource != sampleRateDesired:
        waveformResampled: NDArray[numpy.float32] = resampy.resample(waveform, sampleRateSource, sampleRateDesired, axis=0)
        return waveformResampled
    else:
        return waveform

def writeWav(pathFilename: Union[str, os.PathLike[Any], io.IOBase], waveform: NDArray[Any], sampleRate: int = 44100) -> None:
    """
    Writes a waveform to a WAV file.

    Parameters:
        pathFilename: The path and filename where the WAV file will be saved.
        waveform: The waveform data to be written to the WAV file. The waveform should be in the shape (channels, samples).
        sampleRate (44100): The sample rate of the waveform. Defaults to 44100 Hz.

    Notes:
        The function will create any necessary directories if they do not exist.
        The function will overwrite the file if it already exists without prompting or informing the user.

    Returns:
        None:

    """
    if not isinstance(pathFilename, io.IOBase):
        try:
            pathlib.Path(pathFilename).parent.mkdir(parents=True, exist_ok=True)
        except OSError:
            pass
    soundfile.write(file=pathFilename, data=waveform.T, samplerate=sampleRate, subtype='FLOAT', format='WAV')
